import pytest
from datetime import datetime, date, timezone, timedelta
from test_utils import extend_dict
from fairmeta import RadboudFDP
from dotenv import load_dotenv
import os
import isodate

load_dotenv()

@pytest.fixture(params=["minimal", "full"])
def config(request):
    config = {
        "catalog": {
            "mapping": {
                "challenge_description": ["description"],
                "challenge_title": ["title"]
            }, "contact_point": {
                "hasEmail": "test@testing.com",
                "fn": "David Tester"
            }, "publisher": {
                "mbox": "publisher@publishing.com",
                "identifier": ["identification"],
                "name": ["uitgever"],
                "homepage": "https://uitgeverij.nl"
            }, "license": "cc0",
        
            "dataset": {
                "mapping": {
                    "archive_description": ["description"],
                    "archive_title": ["title"],
                    "challenge_url": ["identifier"],
                    "challenge_keywords": ["keyword"]
                }, "access_rights": "non_public",
                "contact_point": {
                    "hasEmail": "support@test.com",
                    "fn": "testing support"
                }, "creator": {
                    "mbox": "person@testing.com",
                    "identifier": ["test person identifier"],
                    "name": ["datasetmaker"],
                    "homepage": "https://datasetmaker.org"
                }, "publisher": {
                    "mbox": "datapublisher@publishing.com",
                    "identifier": ["identification data"],
                    "name": ["uitgever data"],
                    "homepage": "https://uitgeverij.nl/data"
                }, "keyword": ["Test platform"],
                "theme": ["HEAL"],
                "applicable_legislation": "https://www.legislation.com",
        
                "distribution": {
                    "mapping": {
                        "distribution_access_url": ["access_url"],
                        "distribution_size": ["byte_size"],
                        "distribution_format": ["format"]
                    }, "license": "cc0",
                    "rights": "https://www.example.com/contracts/definitely_a_real_DPA.pdf",
                }
            }
        }
    }
    if request.param == "full":
        extended_config = {
            "catalog": {
                "publisher": {
                    "spatial": ["http://publications.europa.eu/resource/authority/country/NLD"],
                    "publisher_note": "Notitie",
                    "publisher_type": "http://purl.org/adms/publishertype/Academia-ScientificOragnisation",
                },
                "applicable_legislation": "https://www.legislation.com",
                "creator": {
                    "mbox": "catalog@testing.com",
                    "identifier": ["catalogtest person identifier"],
                    "name": ["catalogmaker"],
                    "homepage": "https://catalogmaker.org"
                }, "geographical_coverage": "https://www.geonames.org/countries/NL/the-netherlands.html",
                "homepage": "https://homepage.org",
                "language": "eng",
                "license": "cc0",
                "modification_date": datetime.now(timezone.utc),
                "release_date": datetime.now(timezone.utc),
                "rights": "https://www.websitewithfreetextrights.com",
                # "temporal_coverage": PeriodOfTime(start_date=datetime.now(), end_date=datetime.now()),
                "dataset": {
                    "code_values": "https://www.wikidata.org/wiki/Q32566",
                    "coding_system": "https://www.wikidata.org/wiki/Q81095",
                    "conforms_to": "https://www.wikidata.org/wiki/Q81095",
                    "distribution": {
                        "applicable_legislation": "https://www.legislation.com",
                        "compression_format": "https://www.iana.org/assignments/media-types/application/zip",
                        "description": ["Description of the distribution", "Description in another language"],
                        "documentation": "https://documentation.com",
                        "download_url": "https://google.com",
                        "language": ["Eng", "nld"],
                        "media_type": "https://www.iana.org/assignments/media-types/text/csv",
                        "modification_date": datetime.now(timezone.utc),
                        "packaging_format": "https://package_information.com",
                        "release_date": datetime.now(timezone.utc),
                        "status": "completed",
                        "temporal_resolution": str(timedelta(days=1)),
                        "title": ["title of distribution"]
                    },
                    "frequency": "daily",
                    "purpose": "https://purpose.com",
                    "geographical_coverage": "https://nijmegen.nl",
                    "is_referenced_by": "https://doi.org",
                    "language": "nld",
                    "legal_basis": "InformedConsent",
                    "maximum_typical_age": 55,
                    "minimum_typical_age": 29,
                    "modification_date": datetime.now(timezone.utc),
                    "number_of_records": 99,
                    "number_of_unique_individuals": 88,
                    "personal_data": "https://w3id.org/dpv/pd#Household",
                    "population_coverage": "Adults aged 18–65 diagnosed with type 2 diabetes in the Netherlands between 2015 and 2020",
                    "purpose": "https://w3id.org/dpv#CustomerManagement",
                    "release_date": datetime.now(timezone.utc),
                    "temporal_resolution": isodate.duration_isoformat(timedelta(days=1)),
                    "type": "https://www.type.nl",
                    "status": "withdrawn",
                    "version": "1",
                    "version_notes": ["changed nothing", "still nothing"],
                    "was_generated_by": "https://me.nl"
                }
            }
        }
        config = extend_dict(config, extended_config)

    return config

@pytest.fixture
def api_data():
    return {
        "challenge_description": "Description given by challenge",
        "challenge_title": "Title given by challenge",
        "archive_description": "Description given by archive",
        "archive_title": "Title given by archive",
        "challenge_url": "https://challenge.org/slug",
        "challenge_keywords": ["Medical", "keyword2"],
        "distribution_access_url": "https://testing.com/dist1",
        "distribution_size": 489,
        "distribution_format": "http://publications.europa.eu/resource/authority/file-type/PDF"
    }

@pytest.fixture
def extra_config():
    config = {
        "catalog": {
            "dataset": {
                "keyword": ["CT", "Prostate"],
                "maximum_typical_age": 999,
                "minimum_typical_age": 1,
                "distribution": {
                    "byte_size": 69
                }
            }
        }
    }
    return config

@pytest.fixture
def FDP():
    local_url = "http://localhost:8080"
    FDP = RadboudFDP(token=os.getenv("local_test_FDP_key"))
    FDP.base_url = local_url
    FDP.post_url = local_url
    return FDP
    


from sempyro.dcat import AccessRights
from sempyro.dcat.dcat_dataset import Frequency
from sempyro.hri_dcat import DatasetTheme, DatasetStatus, GeonovumLicences, DistributionStatus

access_rights = {right.name: right for right in AccessRights}

distributionstatuses = {status.name: status for status in DistributionStatus}

frequencies = {frequency.name.lower(): frequency for frequency in Frequency}

licenses = {lic.name: lic.value for lic in GeonovumLicences}

statuses = {status.name: status for status in DatasetStatus}

themes = {theme.name: theme for theme in DatasetTheme}




import pytest
from fairmeta.metadata_model import MetadataRecord
from copy import deepcopy
from pydantic import AnyHttpUrl, Field
from sempyro.hri_dcat import HRICatalog
from rdflib import DCTERMS

class FDPCatalog(HRICatalog):
    is_part_of: [AnyHttpUrl] = Field(
        description="Link to parent object", 
        json_schema_extra={
            "rdf_term": DCTERMS.isPartOf, 
            "rdf_type": "uri"
        })
    
def extend_dict(dictionary: dict, extension: dict) -> dict:
    """Extends a nested dictionary with another nested dictionary preserving structure"""
    for key, value in extension.items():
        if key in dictionary and isinstance(dictionary[key], dict) and isinstance(value, dict):
            extend_dict(dictionary[key], value)
        else:
            dictionary[key] = value
    return dictionary

def extract_paths(config, key):
    """Returns what the path to a field would be if it was in config"""
    path = []
    for k, v in config.items():
        if k == "mapping":
            if key in v:
                return [v[key][0]]
        elif isinstance(v, dict):
            value = extract_paths(v, key)
            if value != []:
                path.append(k)
                path.extend(extract_paths(v, key))
    return path

def resolve_path(obj, path, target, config):
    """Returns the value of a field in config or api_data""" 
    match target:
        case "config":
            for key in path:
                obj = getattr(obj, key)
                if isinstance(obj, list):
                    obj = obj[0]
            return obj
        
        case "api_data":
            key = path[0]
            internal_path = extract_paths(config, key)
            for subkey in internal_path[:-1]:
                obj = getattr(obj, subkey)
                if isinstance(obj, list):
                    obj = obj[0]
            obj = getattr(obj, internal_path[-1])
            return obj, internal_path
        case _:
            raise ValueError

def adapted_instance(target, config, api_data, path, value, extra_config=None):
    """Changes a field in config or api_data and creates an FDPBase with that"""
    match target:
        case "config":
            adapted_data = deepcopy(config) 
        case "api_data":
            adapted_data = deepcopy(api_data) 
        case "multi_conf":
            return MetadataRecord.create_metadata_schema_instance([config, extra_config], api_data)
        case _:
            return MetadataRecord.create_metadata_schema_instance([config], api_data)
    
    d = adapted_data
    for key in path[:-1]:
        d = d[key]
    d[path[-1]] = value
    
    if target == "config":
        return MetadataRecord.create_metadata_schema_instance([adapted_data], api_data)
    else:
        return MetadataRecord.create_metadata_schema_instance([config], adapted_data)

def is_list_field(model: MetadataRecord, path):
    """Helper function to decide if a field should be a list"""
    for key in path[:-1]:
        model = getattr(model, key)
        if isinstance(model, list):
            model = model[0]
    field = model.__class__.model_fields[path[-1]]
    return 'List' in str(field.annotation)


from .metadata_model import MetadataRecord
from .gatherers import GrandChallenge
from .uploader_radboudfdp import RadboudFDP
import argparse
import yaml
import logging
from dotenv import load_dotenv

load_dotenv()

parser = argparse.ArgumentParser()
parser.add_argument('-c', '--config', help="YAML configuration file(s)", action='append', required=True)
parser.add_argument('platform', help="Platform to fetch metadata from")
parser.add_argument('slug', help="Unique identifier of dataset")
parser.add_argument('catalog_name', help="Name of catalog in FDP")
parser.add_argument('--test', action='store_true', help="Run in test mode")
parser.add_argument('-v', '--verbose', action='store_true', help="Verbose logging") 

def main():
    args = parser.parse_args()

    logging.basicConfig(
        level=logging.DEBUG if args.verbose else logging.WARNING,
        format="%(levelname)s: %(message)s"
    )

    logging.info("Loading configuration")
    config_data = []
    for c in args.config:
        with open(c, 'r') as f:
            config_data.append(yaml.safe_load(f))

    platform_key = args.platform.lower()
    configs = []
    for idx, conf in enumerate(config_data):
        platforms = conf.get("platforms", {})
        if platform_key not in platforms:
            raise ValueError(f"Platform '{platform_key}' not found in config file: {args.config[idx]}")
        configs.append(platforms[platform_key])

    logging.info(f"Fetching data from platform: {args.platform}")
    match platform_key:
        case "grand_challenge":
            platform = GrandChallenge()
            api_data = platform.gather_data(f"/{args.slug}")
        case _:
            available = list(config_data[0].get("platforms", {}).keys())
            raise ValueError(f"Unsupported platform: {args.platform}. Pick from: {', '.join(available)}")

    data = MetadataRecord.create_metadata_schema_instance(configs=configs, api_data=api_data)
    logging.info("Validating relaxed metadata schema")
    data.validate()
    MetadataRecord.transform_schema(data)
    logging.info("Validating strict metadata schema")
    data.validate()

    FDP = RadboudFDP(test=args.test)
    FDP.create_and_publish(data, args.catalog_name)

    logging.info("Done")


if __name__=="__main__":
    main()


import pytest
from fairmeta.gatherers.gather_GC import GrandChallenge
from httpx import HTTPStatusError, Request, Response
from gcapi import Client
import test_utils
from fairmeta import metadata_model
import requests

@pytest.mark.parametrize("slug,status_code,exception",[("LUNA16", 200, None),
                                                       ("weird", 404, HTTPStatusError)])
def test_gather_gc_data(monkeypatch, slug, status_code, exception):
    calls = []

    def fake_get(self, path, **kwargs) -> Response:
        calls.append((path, kwargs))
        if status_code == 200:
            content = {"name": slug, "pk": 1}
        else:
            content = {"detail": "Not found"}
        print("path", path)
        return Response(status_code=status_code, json=content, request=Request("GET", f"https://grand-challenge.org/api/v1/challenges{path}"))

    monkeypatch.setattr(Client, "get", fake_get)
    class FakeArchive:
        def __init__(self):
            self.pk = 2
            self.slug = slug.lower()
    
    class FakeArchives:
        def detail(self, slug):
            return FakeArchive()

    class FakeImages:
        def __init__(self):
            self._items = [{"id": "img1"}, {"id": "img2"}]

        def iterate_all(self, params):
            for item in self._items:
                yield item

        def list(self, params):
            return self._items

    def fake_init(self):
        self.client = Client(token="token")
        self.client.archives = FakeArchives()
        self.client.images = FakeImages()

    monkeypatch.setattr("fairmeta.gatherers.gather_GC.GrandChallenge.__init__", fake_init)

    platform = GrandChallenge()
    if exception:
        with pytest.raises(exception) as exc_info:
            platform._gather_challenge(f"{slug}")
        print(exc_info)
    else:
        archive_data = platform._gather_challenge(f"{slug}")
        assert calls
        assert calls[0][0].endswith(f"/challenges/{slug}"), f"Unexpected ending: {calls[0][0]}"
        assert isinstance(archive_data, dict)
        assert archive_data.get("name") == slug


# @pytest.mark.parametrize()
def test_FDP_post_and_publish(FDP, config, api_data):
    schema = test_utils.adapted_instance(None, config, api_data, None, None)
    metadata_model.MetadataRecord.transform_schema(schema)
    catalog_name = "test_catalog"
    urls = FDP.create_and_publish(schema, catalog_name)

    for url in urls:
        rsp = requests.get(url)
        assert rsp.status_code == 200

    FDP.delete(urls[0], confirm=False) # The first URL is the catalog url. Deleting this one deletes all items in the catalog

def test_FDP_update(FDP):
    pass


import os
import gcapi
from urllib.parse import quote

class GrandChallenge:
    def __init__(self, token=None):
        if token:
            self.client = gcapi.Client(token=token)
        else:
            self.client = gcapi.Client(token=os.getenv("GC_key"))

    def gather_data(self, slug) -> dict:
        """Gathers data from challenges and archives based on their slug and combines them in a dictionary"""
        slug = slug.strip("/")
        slug = quote(slug, safe="")

        archive_dict = self._gather_archive(slug)
        challenge_dict = self._gather_challenge(slug)        

        images = self.client.images.iterate_all(params={'archive': archive_dict["pk"]})
        for image in images:
            example_image = image
            break
        num_images = len(self.client.images.list(params={'archive': archive_dict["pk"]}))

        combined_dict = {f'challenge_{k}': v for k, v in challenge_dict.items()}
        combined_dict.update({f'archive_{k}': v for k, v in archive_dict.items()})
        combined_dict.update({'byteSize': num_images})
        return combined_dict
    
    def _gather_archive(self, slug) -> dict:
        archive = self.client.archives.detail(slug=slug.lower()) # Remove lower if you can
        return archive.__dict__
    
    def _gather_challenge(self, slug) -> dict:
        challenge = self.client.get(f"/challenges/{slug}", follow_redirects=True)
        challenge.raise_for_status()
        return challenge.json()


import pytest
from fairmeta import metadata_model, schema_definitions_hri
from fairmeta.metadata_model import MetadataRecord
from pydantic import ValidationError
from sempyro.hri_dcat import HRIAgent, HRIVCard, HRIDataset, HRIDistribution
from sempyro.geo import Location
from rdflib import URIRef
from test_utils import FDPCatalog, resolve_path, adapted_instance, is_list_field

@pytest.mark.parametrize("target,path,value,exception",[(None, None, None, None), # Minimal record
                                                        ("config", ("catalog", "dataset", "contact_point"), "not kind or card", AttributeError),
                                                        ("config", ("catalog", "publisher"), "not agent or HRIAgent", AttributeError),
                                                        ("config", ("catalog", "contact_point", "fn"), "", ValueError), # Empty string in mandatory field
                                                        ("api_data", (["challenge_title"]), [], ValueError), # Empty list in mandatory field
                                                        ("config", ("catalog", "dataset", "contact_point", "fn"), 'something "in quotes" \'excapedquo\'beep', None), # Quotes in string in config file
                                                        ("config", ("catalog", "homepage"), "", None),
                                                        ("config", ("catalog", "contact_point", "hasEmail"), "mailto:email@org.org", None)])
def test_metadata_model_creation(target, config, api_data, path, value, exception):
    """General test for creating a valid instance of the metadata schema"""
    if exception:
        with pytest.raises(exception) as exc_info:
            schema = adapted_instance(target, config, api_data, path, value)
            schema.validate()
            match target:
                case "config":
                    target = resolve_path(schema, path[:-1], target, config)
                    target = getattr(target, path[-1])
                case "api_data":
                    target, _ = resolve_path(schema, path, target, config)
                case _:
                    pass       
            if value:
                assert target == value
            else:
                assert target == None            
        print(exc_info)
        
    else:
        schema = adapted_instance(target, config, api_data, path, value)
        assert isinstance(schema, MetadataRecord)
        extended_keywords = config["catalog"]["dataset"]["keyword"] + api_data["challenge_keywords"]
        assert schema.catalog.dataset.keyword == extended_keywords
        match target:
            case "config":
                target = resolve_path(schema, path[:-1], target, config)
                target = getattr(target, path[-1])
                assert schema.api_data == api_data                
            case "api_data":
                target, _ = resolve_path(schema, path, target, config)
                assert schema.config == [config]
            case _:
                assert schema.config == [config]
                assert schema.api_data == api_data       
        if value:
            assert target == value
        else:
            assert target == None
        schema.validate()
        # Field tests?

@pytest.mark.parametrize("target,path,value,exception",[("config", ("catalog", "contact_point", "fn"),123,ValidationError), 
                                                        ("config", ("catalog", "contact_point", "fn"),True,ValidationError), 
                                                        ("config", ("catalog", "contact_point", "fn"),None,ValueError),
                                                        ("config", ("catalog", "contact_point", "fn"),"Firstname Lastname",None),
                                                        ("api_data", (["archive_title"]), 123, ValidationError),
                                                        ("api_data", (["distribution_size"]), "drie", ValidationError),
                                                        ("api_data", (["archive_title"]), "title", None)])
def test_data_types(target, config, api_data, path, value, exception):
    """Tests if basic types are enforced correctly"""
    schema = adapted_instance(target, config, api_data, path, value)
    if exception:
        with pytest.raises(exception) as exc_info:
            schema.validate() # Bad error: if type is ambigious (Kind | HRIVCard for example) it prints errors for both
        print(exc_info)
    else:
        schema.validate()

@pytest.mark.parametrize("target,path,value,exception,message",[("config", ("catalog", "dataset", "theme"), ["HEAL"], None, None),
                                                        ("config", ("catalog", "dataset", "theme"), ["INVALID_THEME"], ValueError, f"INVALID_THEME incorrect or not supported. Supported values: {', '.join(metadata_model.themes.keys())}"),
                                                        ("config", ("catalog", "dataset", "access_rights"), "public", None, None),
                                                        ("config", ("catalog", "dataset", "access_rights"), "NOT_ALLOWED", ValueError, f"NOT_ALLOWED incorrect or not supported. Supported values: {', '.join(metadata_model.access_rights.keys())}"),
                                                        ("config", ("catalog", "dataset", "theme"), "HEAL", None, None),
                                                        ("config", ("catalog", "dataset", "theme"), "BAD", ValueError, None),
                                                        ("config", ("catalog", "language"), "Eng", None, None),
                                                        ("config", ("catalog", "language"), "En", ValueError, None),
                                                        ("config", ("catalog", "language"), "https://example.com", ValueError, None),
                                                        ("config", ("catalog", "language"), "http://publications.europa.eu/resource/authority/language/anything", None, None), # Not correct but shouldn't raise error
                                                        ("config", ("catalog", "dataset", "frequency"), "quarterly", None, None),
                                                        ("config", ("catalog", "dataset", "frequency"), "dagelijks", ValueError, None),
                                                        ("config", ("catalog", "license"), "cc_byncsa_30", None, None),
                                                        ("config", ("catalog", "license"), "Apache-2.0", ValueError, None),
                                                        ("config", ("catalog", "dataset", "distribution", "status"), "develoP", None, None),
                                                        ("config", ("catalog", "dataset", "distribution", "status"), "asfd", ValueError, None)])
def test_string_to_enum(target, config, api_data, path, value, exception, message):
    """Tests string_to_enum function"""
    schema = adapted_instance(target, config, api_data, path, value)
    if exception:
        if message:
            with pytest.raises(exception, match=message):
                MetadataRecord._string_to_enum(schema)
        else:
            with pytest.raises(exception) as exc_info:
                MetadataRecord._string_to_enum(schema) # Prints good error
            print(exc_info)
    else:
        MetadataRecord._string_to_enum(schema)

@pytest.mark.parametrize("target,path,value,exception",[("config", ("catalog", "contact_point", "hasEmail"), "name", ValueError),
                                                        ("config", ("catalog", "contact_point", "hasEmail"), "email@org.com", None),
                                                        ("config", ("catalog", "dataset", "contact_point", "fn"), "anyone", None),
                                                        ("config", ("catalog", "dataset", "contact_point", "hasUrl"), ["https://example.com"], None),
                                                        ("config", ("catalog", "contact_point", "hasUrl"), ["no url"], ValueError),
                                                        ("config", ("catalog", "contact_point", "hasUrl"), None, None),
                                                        ("config", ("catalog", "contact_point"), [schema_definitions_hri.Kind(hasEmail="email@email.com", fn="name"), HRIVCard(hasEmail="email@email.com", formatted_name="name")], None),
                                                        ("config", ("catalog", "contact_point"), [schema_definitions_hri.Kind(hasEmail="email@email.com", fn="name"), HRIVCard(hasEmail="email@email.com", formatted_name="name"), "random"], ValueError)])
def test_kind_to_hrivcard(target, config, api_data, path, value, exception):
    """Tests the transformation function from Kind to HRIVCard"""
    schema = adapted_instance(target, config, api_data, path, value)
    if exception:
        with pytest.raises(exception) as exc_info:
            MetadataRecord._kind_to_HRIVCard(schema) # Prints good error
        print(exc_info)
    else:
        MetadataRecord._kind_to_HRIVCard(schema)
        target = resolve_path(schema, path[:-1], target, config)
        try:
            assert isinstance(target, HRIVCard)
        except:
            target = getattr(target, path[-1])
            assert all(isinstance(t, HRIVCard) for t in target)

        if path[-1] == "hasUrl" and value is not None:
            if isinstance(target, list):
                for t in target:
                    assert t.contact_page is not None
            else:
                assert target.contact_page is not None

@pytest.mark.parametrize("target,path,value,exception",[("config", ("catalog", "publisher", "mbox"), "name", ValueError), # No email in mailbox
                                                        ("config", ("catalog", "publisher", "homepage"), "no link", ValueError), # No link in homepage
                                                        ("config", ("catalog", "dataset", "creator", "type"), "typen", ValueError), # No link in type
                                                        ("config", ("catalog", "dataset", "creator", "type"), "https://typen.com", None),
                                                        ("config", ("catalog", "dataset", "creator", "spatial"), ["https://Nijmegen.com"], None),
                                                        ("config", ("catalog", "dataset", "creator", "spatial"), ["Nijmegen"], ValueError), # No link in location
                                                        ("config", ("catalog", "dataset", "creator", "spatial"), [Location(geometry="https://Nijmegen.com")], None),                                                    
                                                        ("config", ("catalog", "publisher", "identifier"), ["identification"], None),
                                                        ("config", ("catalog", "publisher", "publisher_type"), None, None),
                                                        ("config", ("catalog", "dataset", "publisher", "publisher_type"), "https://publishertype.com", None),
                                                        ("config", ("catalog", "dataset", "publisher", "publisher_type"), ["https://publishertype.com"], ValueError), # Publisher type in a list when it's not supposed to be
                                                        ("config", ("catalog", "publisher"), [metadata_model.Agent(mbox="dummy@email.com",identifier=["id"],name=["name"],homepage="https://pagina.nl"), HRIAgent(name=["name"],identifier=["id"],mbox="email@email.com",homepage="https://pagina.nl")], None),
                                                        ("config", ("catalog", "publisher"), [metadata_model.Agent(mbox="dummy@email.com",identifier=["id"],name=["name"],homepage="https://pagina.nl"), HRIAgent(name=["name"],identifier=["id"],mbox="email@email.com",homepage="https://pagina.nl"), "random"], ValueError)]) # Not Agent in list
def test_agent_to_hriagent(target, config, api_data, path, value, exception):
    """Tests the tranformation function from Agent to HRIAgent"""
    schema = adapted_instance(target, config, api_data, path, value)
    if exception:
        with pytest.raises(exception) as exc_info:
            MetadataRecord._agent_to_HRIAgent(schema) # Prints good error
        print(exc_info)
    else:
        MetadataRecord._agent_to_HRIAgent(schema)
        target = resolve_path(schema, path[:-1], target, config)
        try:
            assert isinstance(target, HRIAgent) 
        except:
            target = getattr(target, path[-1])
            assert all(isinstance(t, HRIAgent) for t in target)

        if path[-1] == "type":
            assert target.type is not None
        elif path[-1] == "spatial":
            assert target.spatial is not None

@pytest.mark.parametrize("target,path,value,exception,message",[("api_data", (["challenge_title"]), None, ValueError, "Likely put null or null equivalent value in required field"),
                                                        ("api_data", (["challenge_title"]), "title", None, None),
                                                        ("config", ("catalog", "license"), None, None, None),
                                                        ("config", ("catalog", "license"), "cc0", None, None),
                                                        ("config", ("catalog", "dataset", "contact_point", "fn"), None, ValueError, "Likely put null or null equivalent value in required field"),
                                                        ("config", ("catalog", "contact_point", "fn"), "", ValueError, None),
                                                        ("api_data", (["challenge_title"]), [], ValueError, "Likely put null or null equivalent value in required field"),
                                                        ("config", ("catalog", "license"), "", None, None)])
def test_drop_none(target, config, api_data, path, value, exception,message):
    """Tests if drop_none function removes null-equivalent values and doesn't interfere with validation function"""
    schema = adapted_instance(target, config, api_data, path, value)
    if exception:
        if message:
            with pytest.raises(exception, match=message):
                schema.validate()
        else:
            with pytest.raises(exception) as exc_info:
                schema.validate() # Good error
            print(exc_info)
    else:
        schema.validate()
        match target:
            case "config":
                target = resolve_path(schema, path[:-1], target, config)
                target = getattr(target, path[-1])
            case "api_data":
                target, _ = resolve_path(schema, path, target, config)
            case _:
                pass       
        if value:
            assert target == value
        else:
            assert target == None

@pytest.mark.parametrize("target,path,value,exception,message",[("config", ("catalog", "dataset", "applicable_legislation"), "legislature", None, None),
                                                        ("config", ("catalog", "dataset", "applicable_legislation"), ["https://license.com"], None, None),
                                                        ("config", ("catalog", "dataset", "purpose"), None, None, None),
                                                        ("config", ("catalog", "dataset", "purpose"), "purposefield", None, None),
                                                        ("config", ("catalog", "dataset", "purpose"), ["purpose field", "purpose_2"], None, None),
                                                        ("api_data", (["challenge_url"]), ["idee"], None, None), # Warning?
                                                        ("api_data", (["challenge_url"]), ["idee2", "illegal_id"], TypeError, "Found list where it is not supposed to be: identifier")])
def test_ensure_lists(target, config, api_data, path, value, exception, message):  
    """Tests the function that creates lists where it needs to be, and removes lists where they don't need to be"""  
    schema = adapted_instance(target, config, api_data, path, value)
    if exception:
        if message:
            with pytest.raises(exception, match=message):
                MetadataRecord._ensure_lists(schema)
        else:
            with pytest.raises(exception) as exc_info:
                MetadataRecord._ensure_lists(schema) # Prints good error
            print(exc_info)
    else:
        MetadataRecord._ensure_lists(schema)
        match target:
            case "config":
                target = resolve_path(schema, path[:-1], target, config)
                target = getattr(target, path[-1])
                list_type = is_list_field(schema, path)
            case "api_data":
                target, internal_path = resolve_path(schema, path, target, config)
                list_type = is_list_field(schema, internal_path)
            case _:
                pass
        if list_type and value is not None:
            assert isinstance(target, list)
        else:
            assert not isinstance(target, list)

@pytest.mark.parametrize("target,path,value,exception",[(None, None, None, None),])
def test_transformation_hri(target, config, api_data, path, value, exception):
    """Tests transformation from unprocessed MetadataRecord to HRI schema"""
    schema = adapted_instance(target, config, api_data, path, value)
    schema.transform_schema()
    disallowed_fields = {"distribution", "dataset"}
    filtered_fields = {k: v for k, v in vars(schema.catalog).items() if k not in disallowed_fields and v is not None}
    catalog = FDPCatalog(
        is_part_of=[URIRef("https://test.com")],
        dataset=[],
        **filtered_fields)
    for dataset in schema.catalog.dataset:
        filtered_fields = {k: v for k, v in vars(dataset).items() if k not in disallowed_fields and v is not None}
        hri_dataset = HRIDataset(
            **filtered_fields
        )
        for distribution in dataset.distribution:
            filtered_fields = {k: v for k, v in vars(distribution).items() if k not in disallowed_fields and v is not None}
            hri_distribution = HRIDistribution(
                **filtered_fields
            )

@pytest.mark.parametrize("target,path,value",[("multi_conf", None, None)])
def test_extra_configs(target, config, api_data, path, value, extra_config):
    """Tests if multiple config files are handled correctly"""
    schema = adapted_instance(target, config, api_data, path, value, extra_config)
    assert schema.catalog.dataset.keyword == ["Test platform", "CT", "Prostate", "Medical", "keyword2"]
    assert schema.catalog.dataset.maximum_typical_age == extra_config["catalog"]["dataset"]["maximum_typical_age"]


import os
import requests
from urllib.parse import urlparse, urlunparse
from .metadata_model import MetadataRecord
from pydantic import AnyHttpUrl, Field
from sempyro.hri_dcat import HRICatalog, HRIDataset, HRIDistribution
from rdflib import DCTERMS, URIRef
import logging

class FDPCatalog(HRICatalog):
    is_part_of: [AnyHttpUrl] = Field(
        description="Link to parent object", 
        json_schema_extra={
            "rdf_term": DCTERMS.isPartOf, 
            "rdf_type": "uri"
        })

class RadboudFDP:
    def __init__(self, test=False, token=None):
        self.test = test
        if token:
            self.FDP_key = token
        else:
            self.FDP_key = os.getenv("Radboud_FDP_key")
        self.base_url = "https://fdp.radboudumc.nl"
        if test:
            self.post_url = "https://fdp.radboudumc.nl/acc"
        else:
            self.post_url = self.base_url
        

    def create_and_publish(self, metadata_record: MetadataRecord, catalog_name: str) -> list[str]:
        """Uploads a MetadataRecord object to Radboud FDP and returns url's"""
        urls = []
        disallowed_fields = {"distribution", "dataset"}
        filtered_fields = {k: v for k, v in vars(metadata_record.catalog).items() if k not in disallowed_fields and v is not None}
        catalog = FDPCatalog(
            is_part_of=[URIRef(self.base_url)],
            dataset = [],
            **filtered_fields
        )
        metadata_catalog_record = catalog.to_graph(URIRef(f"{self.post_url}/catalog/{catalog_name}"))
        metadata_catalog_turtle = metadata_catalog_record.serialize(format="turtle")
        post_rsp = self._post(metadata_catalog_turtle, "catalog")
        metadata_catalog_url = post_rsp.headers["Location"]
        urls.append(metadata_catalog_url)

        for dataset in metadata_record.catalog.dataset:
            filtered_fields = {k: v for k, v in vars(dataset).items() if k not in disallowed_fields and v is not None}
            hri_dataset = HRIDataset(
                **filtered_fields
            )
            metadata_dataset_record = hri_dataset.to_graph(subject=URIRef(hri_dataset.identifier))
            metadata_dataset_record.add((URIRef(hri_dataset.identifier), DCTERMS.isPartOf, URIRef(metadata_catalog_url)))
            metadata_dataset_turtle = metadata_dataset_record.serialize(format="turtle")
            post_rsp = self._post(metadata_dataset_turtle, "dataset")
            metadata_dataset_url = post_rsp.headers["Location"]
            urls.append(metadata_dataset_url)

            # Cannot test this due to SHACLs: byteSize gives DatatypeConstraintComponent* and title gives MinCountConstraintComponent (Even though it is not mandatory)
            # in SeMPyRO, byteSize is xsd:integer, I think defining it as xsd:nonnegativeinteger would immediately solve this problem.
            # for distribution in dataset.distribution:
            #     filtered_fields = {k: v for k, v in vars(distribution).items() if k not in disallowed_fields and v is not None}
            #     hri_distribution = HRIDistribution(
            #         **filtered_fields
            #     )
            #     access_url_str = str(hri_distribution.access_url)
            #     distribution_uri = URIRef(f"{hri_dataset.identifier}/distribution/{access_url_str.split('/')[-1]}")
            #     metadata_distribution_record = hri_distribution.to_graph(subject=distribution_uri)
            #     metadata_distribution_record.add((distribution_uri, DCTERMS.isPartOf, URIRef(f"{metadata_dataset_url}")))
            #     metadata_distribution_turtle = metadata_distribution_record.serialize(format="turtle")

            #     post_rsp = self._post(metadata_distribution_turtle, "distribution")
            #     metadata_distribution_url = post_rsp.headers["Location"]
            #     urls.append(metadata_distribution_url)

            #     publish_rsp = self._publish(metadata_distribution_url)

            publish_rsp = self._publish(metadata_dataset_url)

        publish_rsp = self._publish(metadata_catalog_url)

        return urls


    def update(self, target: str, metadata_record: MetadataRecord, url: str, pointer_url):
        disallowed_fields = {"distribution", "dataset"}            
        match target:
            case "catalog":
                filtered_fields = {k: v for k, v in vars(metadata_record).items() if k not in disallowed_fields and v is not None}
                catalog = FDPCatalog(
                    is_part_of=[URIRef(self.base_url)],
                    dataset = [pointer_url],
                    **filtered_fields
                )
                metadata_catalog_record = catalog.to_graph(URIRef(url))
                metadata_catalog_turtle = metadata_catalog_record.serialize(format="turtle")
                rsp = self._put(metadata_catalog_turtle, url)

            case "dataset":
                filtered_fields = {k: v for k, v in vars(metadata_record).items() if k not in disallowed_fields and v is not None}
                hri_dataset = HRIDataset(
                    **filtered_fields
                )
                metadata_dataset_record = hri_dataset.to_graph(subject=URIRef(url))
                metadata_dataset_record.add((URIRef(url), DCTERMS.isPartOf, URIRef(pointer_url)))
                metadata_dataset_turtle = metadata_dataset_record.serialize(format="turtle")
                rsp = self._put(metadata_dataset_turtle, url)

            case _:
                raise ValueError(f"Target: {target} invalid")

        logging.info(f"Updating: {target}, response (should be 200): {rsp}")
        return rsp
    

    def delete(self, url :str, confirm: bool=True):
        if confirm:
            while True:
                user_input = input(f"Are you sure you want to DELETE {url}? [yes/no]: ").strip().lower()
                if user_input == "yes":
                    break
                elif user_input in ("", "n", "no"):
                    logging.info(f"Skipped deletion of {url}")
                    return None
                else:
                    print("Type 'yes' to confirm deletion or 'no' to cancel.")

        headers = {
            'Authorization': f'Bearer {self.FDP_key}',
        }
        rsp = requests.delete(url, headers=headers)
        self._check_response(rsp, action="DELETE")
        logging.info(f"Deleting: {url}, response (should be 204): {rsp}")
        return rsp


    def _post(self, turtle, location) -> str:
        url = f"{self.post_url}/{location}"
        headers = {
            'Authorization': f'Bearer {self.FDP_key}',
            'Content-Type': 'text/turtle'
        }
        rsp = requests.post(url, headers=headers, data=turtle, allow_redirects=True)
        self._check_response(rsp, action="POST")
        logging.info(f"Posting: {location}, response (should be 201): {rsp}")
        return rsp
    

    def _publish(self, url):
        if self.test:
            parsed = urlparse(url)
            new_path = "/acc" + parsed.path
            url = urlunparse(parsed._replace(path=new_path))

        publish_url = f"{url}/meta/state"
        headers = {
            'Authorization': f'Bearer {self.FDP_key}',
            'Accept': 'application/json',
            'Content-Type': 'application/json'
        }
        json_data = {
            'current': 'PUBLISHED'
        }
        rsp = requests.put(url=publish_url, headers=headers, json=json_data)
        self._check_response(rsp, action="PUT")
        logging.info(f"Published, this should be 200: {rsp}")
        return rsp

    
    def _put(self, turtle, url):
        if self.test:
            parsed = urlparse(url)
            new_path = "/acc" + parsed.path
            url = urlunparse(parsed._replace(path=new_path))

        headers = {
                'Authorization': f'Bearer {self.FDP_key}',
                'Accept': 'text/turtle',
                'Content-Type': 'text/turtle'
            }
        
        rsp = requests.put(url, headers=headers, data=turtle)
        self._check_response(rsp, action="PUT")
        return rsp
    
    
    def _check_response(self, rsp: requests.Response, action: str = "request"):
        """Raise detailed error if HTTP response indicates failure."""
        try:
            rsp.raise_for_status()
        except requests.exceptions.HTTPError as e:
            logging.error(f"{action.capitalize()} failed with status {rsp.status_code} for {rsp.url}")
            logging.error(f"Response headers: {rsp.headers}")
            logging.error(f"Response text:\n{rsp.text}")
            raise

        



from .schema_definitions_hri import Agent, Catalog, Dataset, Distribution, Kind
import logging
from pydantic import BaseModel, AnyHttpUrl, TypeAdapter, ValidationError
from sempyro.hri_dcat import HRIVCard, HRIAgent
from typing import List, Optional
from .mappings import themes, access_rights, frequencies, statuses, licenses#, distributionstatuses
import warnings

def _is_valid_http_url(value: str) -> bool:
    try:
        TypeAdapter(AnyHttpUrl).validate_python(value)
        return True
    except ValidationError:
        return False

class MetadataRecord(BaseModel):
    catalog: Catalog
    config: Optional[List[dict]] = None
    api_data: Optional[dict] = None

    @classmethod
    def create_metadata_schema_instance(cls, configs : list = None, api_data : dict = None) -> "MetadataRecord":
        """Fills the metadata schema using configs and API data"""
        schema_obj = cls.model_construct(config=configs, api_data=api_data)
        if schema_obj.config is not None:
            if not isinstance(schema_obj.config, list):
                schema_obj.config = [schema_obj.config]
            for config in schema_obj.config:
                MetadataRecord._fill_fields_default(schema_obj, config)
            if schema_obj.api_data is not None:
                for config in schema_obj.config:
                    MetadataRecord._populate_schema(schema_obj, api_data, config)
        return schema_obj
    
    def transform_schema(self):
        """Calls all functions to change fields to Health-RI complient formats"""
        MetadataRecord._ensure_lists(self)
        MetadataRecord._string_to_enum(self)
        MetadataRecord._agent_to_HRIAgent(self)
        MetadataRecord._kind_to_HRIVCard(self)  

    def validate(self):
        """Validates if mandatory fields have acceptable values"""
        cleaned = MetadataRecord._drop_none(self)
        type(self).model_validate(cleaned, strict=True)
        logging.info("Validation successful")

    # def validate(self): 
    #     self.model_validate(self.model_dump()) 
    #     logging.info("Validation successful") 
          
    @staticmethod
    def _fill_fields_default(schema_obj, config: dict):
        """Recursively fills in the fields from the config file"""
        try:
            for key, value in config.items():
                if isinstance(value, list):
                    match key:
                        case "keyword":
                            setattr(schema_obj, key, getattr(schema_obj, key) + value)
                        case _:
                            setattr(schema_obj, key, value)
                else:
                    match key:
                        case "catalog":
                            if not hasattr(schema_obj, key):
                                setattr(schema_obj, key, Catalog.model_construct())
                            MetadataRecord._fill_fields_default(getattr(schema_obj, key), value)
                        case "dataset":
                            if not hasattr(schema_obj, key):
                                setattr(schema_obj, key, Dataset.model_construct())
                            MetadataRecord._fill_fields_default(getattr(schema_obj, key), value)
                        case "distribution":
                            if not hasattr(schema_obj, key) or getattr(schema_obj, key) == None:
                                setattr(schema_obj, key, Distribution.model_construct())
                            MetadataRecord._fill_fields_default(getattr(schema_obj, key), value)
                        case "creator" | "publisher":
                            if not hasattr(schema_obj, key) or getattr(schema_obj, key) == None:
                                setattr(schema_obj, key, Agent.model_construct())
                            MetadataRecord._fill_fields_default(getattr(schema_obj, key), value)
                        case "contact_point":
                            if not hasattr(schema_obj, key) or getattr(schema_obj, key) == None:
                                setattr(schema_obj, key, Kind.model_construct())
                            MetadataRecord._fill_fields_default(getattr(schema_obj, key), value)
                        case "mapping":
                            pass
                        case _:
                            if value:
                                try:
                                    v = getattr(schema_obj, key)
                                    if v == None:
                                        raise AttributeError
                                    else:
                                        warnings.warn(f"Field value overwritten: {key}: {getattr(schema_obj, key)} with {value}")
                                        raise AttributeError
                                except AttributeError:
                                    setattr(schema_obj, key, value)

        except AttributeError as e:
            print("Likely in one of the fields creator, publisher, or contact_point, something else than a dictionary or list was given")
            raise e


    @staticmethod
    def _populate_schema(schema_obj, api_data: dict, config: dict):
        """Recursively fills in the fields from the api data"""
        for field, value in config.items():
            match field:
                case "catalog":
                    MetadataRecord._populate_schema(getattr(schema_obj, field), api_data, value)
                case "dataset":
                    MetadataRecord._populate_schema(getattr(schema_obj, field), api_data, value)
                case "distribution":
                    MetadataRecord._populate_schema(getattr(schema_obj, field), api_data, value)
                case "mapping":
                    if isinstance(value, dict):
                        for api_field, internal_fields in value.items():
                            if api_field in api_data:
                                for internal_field in internal_fields:
                                    if api_data[api_field]:
                                        if internal_field == "keyword" and isinstance(schema_obj.keyword, list):
                                            setattr(schema_obj, internal_field, schema_obj.keyword + api_data[api_field]) # Not using extend here because it changes keyword in config
                                        else:
                                            setattr(schema_obj, internal_field, api_data[api_field])


    @staticmethod
    def _ensure_lists(schema_obj):
        """Changes all fields that need to be lists in the Health-RI metadata schema into lists, and ensures fields that are not allowed to be lists are not"""
        for field_name, field in schema_obj.model_fields.items():
            value = getattr(schema_obj, field_name)
            if isinstance(value, BaseModel):
                MetadataRecord._ensure_lists(value)
                
            is_list_type = 'List' in str(field.annotation)
            if is_list_type and not isinstance(value, list) and value is not None:
                setattr(schema_obj, field_name, [value])
            elif not is_list_type and isinstance(value, list):
                if len(value) == 1:
                    setattr(schema_obj, field_name, value[0])
                    warnings.warn(f"Please do not put list in field: {field_name}")
                else:
                    raise TypeError(f"Found list where it is not supposed to be: {field_name}")


    @staticmethod
    def _string_to_enum(schema_obj):
        """Changes field values into Health-RI supported categories"""
        for field_name, _ in schema_obj.model_fields.items():
            value = getattr(schema_obj, field_name)
            if value:
                if isinstance(value, BaseModel):
                    MetadataRecord._string_to_enum(value)

                elif isinstance(value, list):
                    for v in value:
                        if isinstance(v, BaseModel):
                            MetadataRecord._string_to_enum(v)

                dict_backed = {
                    "access_rights": access_rights,
                    "theme": themes,
                    "license": licenses,
                    "status": statuses,
                    "frequency": frequencies,
                }
                transformer_backed = ["format", "language", "legal_basis", "personal_data", "purpose"]

                if field_name in dict_backed:
                    kind = dict_backed[field_name]
                    if isinstance(value, list):
                        for i, v in enumerate(value):
                            value[i] = MetadataRecord._to_enum(v, kind)
                    else:
                        setattr(schema_obj, field_name, MetadataRecord._to_enum(value, kind))

                elif field_name in transformer_backed:
                    if isinstance(value, list):
                        for i, v in enumerate(value):
                            value[i] = MetadataRecord._to_enum(v, field_name)
                    else:
                        setattr(schema_obj, field_name, MetadataRecord._to_enum(value, field_name))

                elif field_name == "spatial":
                    pass

                else:
                    pass


    @staticmethod
    def _to_enum(value, kind):
        match kind:
            case "format":
                return MetadataRecord._format_transformation(value)
            case "language":
                return MetadataRecord._language_transformation(value)
            case "legal_basis":
                return MetadataRecord._legal_basis_transformation(value)
            case "personal_data":
                return MetadataRecord._personal_data_transformation(value)
            case "purpose":
                return MetadataRecord._purpose_transformation(value)
            case _:
                try:
                    return kind[value.lower()]
                except:
                    if not value in kind.values():
                        raise ValueError(f"{value} incorrect or not supported. Supported values: {', '.join(kind.keys())}")
        

    @staticmethod
    def _format_transformation(value):
        if not _is_valid_http_url(value):
            return f"http://publications.europa.eu/resource/authority/file-type/{value}"
        else:
            if not "http://publications.europa.eu/resource/authority/file-type/" in value:
                raise ValueError(f"Format should be in the form: http://publications.europa.eu/resource/authority/file-type/<code> not {value}")
            else:
                return value


    @staticmethod
    def _language_transformation(value):
        if not _is_valid_http_url(value):
            match value.lower():
                case "nederlands" | "dutch" | "nld":
                    return "http://publications.europa.eu/resource/authority/language/NLD"
                case "english" | "engels" | "eng":
                    return "http://publications.europa.eu/resource/authority/language/ENG"
                case _:
                    raise ValueError("For language: either provide 'nld' or 'eng', or in the form http://publications.europa.eu/resource/authority/language/<code>")
        else:
            if not "http://publications.europa.eu/resource/authority/language/" in value:
                raise ValueError(f"Language should be in the form: http://publications.europa.eu/resource/authority/language/<code> not {value}")
            else:
                return value
    
    
    @staticmethod
    def _legal_basis_transformation(value):
        if not _is_valid_http_url(value):
            return f"https://w3id.org/dpv#{value}"
        else:
            return value
        
    @staticmethod
    def _personal_data_transformation(value):
        if not _is_valid_http_url(value):
            return f"https://w3id.org/dpv/pd#{value}"
        else:
            return value
        
    @staticmethod
    def _purpose_transformation(value):
        if not _is_valid_http_url(value):
            return f"https://w3id.org/dpv#{value}"
        else:
            return value

    @staticmethod
    def _agent_to_HRIAgent(schema_obj):
        """Changes Agents into Health-RI Agents"""
        for field_name, _ in schema_obj.model_fields.items():
            value = getattr(schema_obj, field_name)
            if isinstance(value, Agent):
                setattr(schema_obj, field_name, MetadataRecord._create_HRIAgent(value)) 
                
            elif isinstance(value, list) and any(isinstance(v, Agent) for v in value):
                new_agents = []
                for agent in value:
                    if isinstance(agent, Agent):
                        new_agents.append(MetadataRecord._create_HRIAgent(agent))
                    elif isinstance(agent, HRIAgent):
                        new_agents.append(agent)
                    else:
                        raise ValueError("Encountered not Agent or HRIAgent in list")
                setattr(schema_obj, field_name, new_agents)

            elif isinstance(value, BaseModel):
                MetadataRecord._agent_to_HRIAgent(value)    
            elif isinstance(value, list):
                for v in value:
                    if isinstance(v, BaseModel):
                        MetadataRecord._agent_to_HRIAgent(v)

    @staticmethod
    def _create_HRIAgent(agent: Agent) -> HRIAgent:
        kwargs = {
            'mbox': agent.mbox,
            'identifier': agent.identifier,
            'name': agent.name,
            'homepage': agent.homepage
        }
        if agent.spatial is not None:
            kwargs['spatial'] = agent.spatial
        if agent.type is not None:
            kwargs['type'] = agent.type
        if agent.publisher_type is not None:
            kwargs['publisher_type'] = agent.publisher_type
        if agent.publisher_note is not None:
            kwargs['publisher_note'] = agent.publisher_note
        return HRIAgent(**kwargs)

    @staticmethod
    def _kind_to_HRIVCard(schema_obj):
        """Changes kinds into Health-RI VCards"""
        for field_name, _ in schema_obj.model_fields.items():
            value = getattr(schema_obj, field_name)
            if isinstance(value, Kind):
                setattr(schema_obj, field_name, MetadataRecord._create_HRIVCard(value))
            elif isinstance(value, list) and any(isinstance(v, Kind) for v in value):
                new_card = []
                for kind in value:
                    if isinstance(kind, Kind):
                        new_card.append(MetadataRecord._create_HRIVCard(kind))
                    elif isinstance(kind, HRIVCard):
                        new_card.append(kind)
                    else:
                        raise ValueError("Encountered not Kind or VCard in list")
                setattr(schema_obj, field_name, new_card)

            elif isinstance(value, BaseModel):
                MetadataRecord._kind_to_HRIVCard(value)

            elif isinstance(value, list):
                for v in value:
                    if isinstance(v, BaseModel):
                        MetadataRecord._kind_to_HRIVCard(v)

    @staticmethod
    def _create_HRIVCard(kind: Kind) -> HRIVCard:
        kwargs = {
            'hasEmail': kind.hasEmail,
            'formatted_name': kind.fn
        }
        if kind.hasUrl is not None:
            kwargs['contact_page'] = kind.hasUrl
        return HRIVCard(**kwargs)

    # The _drop_none function below is necessary because when validating an HRIVCard or HRIAgent which has
    # optional values that are None, it gives a ValidationError
    @staticmethod
    def _drop_none(data):
        """Removes all None values in non mandatory fields"""
        if isinstance(data, BaseModel):
            result = {}
            for name, field in data.model_fields.items():
                try:
                    value = getattr(data, name)
                except:
                    raise ValueError("Likely put null or null equivalent value in required field")
                if value is not None or field.is_required():
                    result[name] = MetadataRecord._drop_none(value)
            return result

        elif isinstance(data, dict):
            return {k: MetadataRecord._drop_none(v) for k, v in data.items() if v is not None}
        elif isinstance(data, list):
            return [MetadataRecord._drop_none(v) for v in data if v is not None]
        else:
            return data